"""
Uses the parsed GDScriptClasses dictionary parsed from the JSON
data to output the reStructuredText documents
"""

import json
import re
from argparse import Namespace
from typing import List

from .config import LOGGER
from .gdscript_objects import (Element, GDScriptClass, GDScriptClasses,
                               ProjectInfo)
from .make_restructured import (RestructuredDocument, RestructuredSection,
                                make_bold, make_code_block, make_comment,
                                make_heading, make_link, make_table_header,
                                make_table_row, surround_with_html,
                                wrap_in_newlines, make_prop_table,
                                make_element, make_func_table)
from .utils import ratify_class_name


def convert_to_restructured(
    classes: GDScriptClasses, arguments: Namespace, info: ProjectInfo
) -> List[RestructuredDocument]:
    """
    Takes a list of dictionaries, each one representing a GDScript class
    and  converts it to reStructuredText.  It returns a list of
    reStructuredText documents.
    """
    restructured: List[RestructuredDocument] = []
    if arguments.make_index:
        restructured.append(_write_index_page(classes, info))
    for entry in classes:
        restructured.append(_as_restructured(classes, entry, arguments))
    return restructured

def _as_restructured(
    classes: GDScriptClasses, gdscript: GDScriptClass, arguments: Namespace
) -> RestructuredDocument:
    """
    Converts the data from a GDScript class into reStructuredText.
    """

    content: List[str] = []

    name: str = gdscript.name

    doc_ref: str = "class_" + name
    
    if "abstract" in gdscript.metadata.tags:
        name += " " + surround_with_html("(abstract)", "small")

    content += [
        make_comment(
            "Auto-generated from JSON by GDScript restructured maker.\n"
            "Do not edit this document directly as all changes will be\n"
            "overwritten on the next auto-generation." 
        )
        + "\n\n"
        + ".. _{}:".format(doc_ref) 
    ]

    content += [*make_heading(name, 1)]
    if gdscript.extends:
        extends_list: List[str] = gdscript.get_extends_tree(classes)
        extends_links = [make_link(entry) for entry in extends_list]
        content += [make_bold("Extends:") + " " + " < ".join(extends_links)]
        description = _replace_references(classes, gdscript, gdscript.description)
        content += [*RestructuredSection("Description", 2, [description]).as_text()]

    content += _write_class(classes, gdscript, 2)
    # if gdscript.signals:
    #     content += RestructuredSection(
    #         "Signals", 2, _write_signals(classes, gdscript)
    #     ).as_text()

    if gdscript.sub_classes:
        content += make_heading("Sub-classes", 2)
    for cls in gdscript.sub_classes:
        content += _write_class(classes, cls, 3, True)

    return RestructuredDocument(gdscript.name, doc_ref, content)

    
    
def _write_class(
    classes: GDScriptClasses,
    gdscript: GDScriptClass,
    heading_level: int,
    is_inner_class: bool = False,
) -> List[str]:
    restructured: List[str] = []
    if is_inner_class:
        restructured += make_heading(gdscript.name, heading_level)
    for attribute, title, table in [
        ("members", "Properties", True),
        ("functions", "Methods", True),
        ("signals", "Signals", False),
        ("enums", "Enumerations", False),
        ("constants", "Constants", False),
        ("members", "Property Descriptions", False),
        ("functions", "Method Descriptions", False),
    ]:
        if not getattr(gdscript, attribute):
            continue
        restructured += RestructuredSection(
            title,
            heading_level + 1 if is_inner_class else heading_level,
            _write(attribute, classes, gdscript, table),
        ).as_text()
    return restructured




def _writ_summary(gdscript: GDScriptClass, key: str) -> List[str]:
    element_list = getattr(gdscript, key)
    if not element_list:
        return []
    restructured: List[str] = make_table_header(["Type", "Name"])
    return restructured + [make_table_row(item.summerize()) for item in element_list]


def _write(
    attribute: str,
    classes: GDScriptClasses,
    gdscript:GDScriptClass,
    table: bool,
    heading_level: int = 3
) -> List[str]:
    assert hasattr(gdscript, attribute)
    
    restructured: List[str] = []
    element = getattr(gdscript, attribute)
    if table:
        if attribute == "members":
            restructured.extend(make_prop_table(element, gdscript.name))
        else:
            restructured.extend(make_func_table(element, gdscript.name))
    else:
        restructured.extend(make_element(attribute, element, gdscript.name))
        # for element in getattr(gdscript, attribute):
        #     # restructured.extend(make_heading(element.get_heading_as_string(), heading_level))
        #     restructured.extend([make_code_block(element.signature), ""])
        #     restructured.extend(element.get_unique_attributes_as_restructured())
        #     restructured.append("")
        #     description: str = _replace_references(classes, gdscript, element.description)
        #     restructured.append(description)

    return restructured

def _write_signals(classes:GDScriptClasses, gdscript: GDScriptClass) -> List[str]:
    return wrap_in_newlines(
        [
            "- **{}**\n\n{}".format(
                s.signature.split(' ', 1)[1],
                _replace_references(classes, gdscript, s.description)
            )
            for s in gdscript.signals
        ]
    )


def _write_index_page(classes: GDScriptClasses, info: ProjectInfo) -> RestructuredDocument:
    title: str = "{}".format(info.name)
    version: str = "Version: {}".format(info.version)
    content: List[str] =[
        *RestructuredSection(title, 1, [version] + [""] + [info.description]).as_text(),
        *RestructuredSection("Contents", 2, _write_table_of_contents(classes)).as_text()
    ]
    return RestructuredDocument("index", "", content)

def _write_table_of_contents(classes: GDScriptClasses) -> List[str]:
    toc: List[str] = [
        "..  toctree::",
        "    :maxdepth: 1",
        "    :caption: API",
        "    :glob:",
        "\n"
    ]

    for c in classes.class_index:
        index_name: str = ratify_class_name(c)
        LOGGER.info( "Adding index " + index_name)
        toc.append("    " + index_name)

    return toc


def _replace_references(
    classes: GDScriptClasses, gdscript: GDScriptClass, description: str
) -> str:
    """
    Finds and replaces references to other classes or methods in the description.
    """
    ERROR_MESSAGE = {
        "class": "Class {} not found in the class index.",
        "member": "Symbol {} not found in {}.  The name might be incorrect.",
    }
    ERROR_TAIL = "The name might be incorrect."

    references: re.Match = re.findall(r"\[,+\]", description)
    for reference in references:
        # Matches [ClassName], [symbol] and [ClassName.symbol]
        match: re.Match = re.match(
            r"\[([A-Z][a-zA-Z0-9]*)?\.?([a-z0-9_]+)?\]", reference
        )
        if not match:
            continue

        class_name, member = match[1], match[2]

        if class_name and class_name not in classes.class_index:
            LOGGER.warning(ERROR_MESSAGE["class"].format(class_name) + ERROR_TAIL)
            continue

        if member and class_name:
            if member not in classes.class_index[class_name]:
                LOGGER.warning(
                    ERROR_MESSAGE["member"].format(member, gdscript.name) +ERROR_TAIL
                )
                continue

        display_text, path = "", "../"
        if class_name:
            display_text, path = class_name, class_name
        if class_name and member:
            display_text+= "."
            path += "/"
        if member:
            display_text += member
            path += "#" + member.replace("_", "-")

        link: str = make_link(display_text, path)
        description = description.replace(reference,link, 1)
    return description






"""Generic utility functions for GDScript docs maker."""


def build_re_pattern(tag_name: str) -> str:
    """Returns a string pattern to match for JSDoc-style tags, with the form @tag_name
    or @tag_name - value
    The match pattern has a match group for the value."""
    return "^@{} ?-? ?(.+)?".format(tag_name)


def ratify_class_name(class_name: str) -> str:
    """Given a class name with the possible extraneous extensions 
    returns the name without the extensions and converted to lower case
    """

    return class_name.split(".", 1)[0]


"""
Merges JSON dumped by Godot's script language server and converts
it to a rest document formatted in the look and feel of the Godot
API documentation
"""

import json
import logging
import os
import sys
from argparse import Namespace
from itertools import repeat
from typing import List

from .src import command_line
from .src.config import LOG_LEVELS, LOGGER
from .src.gdscript_objects import GDScriptClasses, ProjectInfo
from .src.make_restructured import RestructuredDocument
from .src.convert_to_restructured import convert_to_restructured
from .src.utils import ratify_class_name


def main():
    args: Namespace = command_line.parse()

    if args.version:
        print("GDScript REST Doc Maker version 0.1.1")
        sys.exit()
    
    logging.basicConfig(level = LOG_LEVELS[min(args.verbose, len(LOG_LEVELS) - 1)])
    json_files: List[str] = [f for f in args.files if f.lower().endswith(".json")]
    LOGGER.info("Processing JSON files: {}".format(json_files))
    for f in json_files:
        with open(f,"r") as json_file:
            data: list = json.loads(json_file.read())
            
            if data["version"] == None:
                LOGGER.info(
                    "Project has no version number - setting it to {}".format(args.doc_version)
                )
                data["version"] = args.doc_version

            project_info: ProjectInfo = ProjectInfo.from_dict(data)

            classes: GDScriptClasses = GDScriptClasses.from_dict_list(data["classes"])
            classes_count: int = len(classes)

            LOGGER.info(
                "Project {}, version {}".format(project_info.name, project_info.version)
            )

            LOGGER.info(
                "Processing {} classes in {}".format(classes_count, os.path.basename(f))
            )

            documents: List[RestructuredDocument] = convert_to_restructured(
                classes, args, project_info
            )

            if args.dry_run:
                LOGGER.debug("Generated {} reStructured documents.".format(len(documents)))
                list(map(lambda doc: LOGGER.debug(doc), documents))
            else:
                if not os.path.exists(args.path):
                    LOGGER.info("Creating directory " + args.path)
                    os.mkdir(args.path)
                
                LOGGER.info(
                    "Saving {} reStructured files to {}".format(len(documents), args.path)
                )
                list(map(save, documents, repeat(args.path)))


def save(
    document: RestructuredDocument,
    dirpath: str
):
    path: str = os.path.join(dirpath, document.get_filename())
    with open(path, "w") as file_out:
        LOGGER.debug("Saving reStructured file " + path)
        file_out.write(document.as_string())

            
if __name__== "__main__":
    main()



"""General functions and utilities to write reStructured documents.
"""
from argparse import RawDescriptionHelpFormatter
from os import name
import re
import json
from dataclasses import dataclass
from typing import Dict, List, Any
import logging
from .config import LOG_LEVELS, LOGGER
from .utils import ratify_class_name

api_ref = {}

with open("godot_api_calls.json", "r") as api_json:
    api_ref: list = json.loads(api_json.read())
    LOGGER.debug(
        "api_ref is {}".format(type(api_ref))
    )
    
@dataclass
class RestructuredDocument:
    title: str
    doc_ref: str
    content: List[str]

    def get_filename(self):
        return ratify_class_name(self.title) + ".rst"

    def as_string(self) -> str:
        """
        Removes duplicate empty lines from the document and returns it as a
        string.
        """
        text: str = "\n".join(self.content)
        return re.sub(r"\n\n+", "\n\n", text)

    def __repr__(self):
        return "RestructuredDocument(title={}, doc_ref={} content={})".format(
            self.title, self.doc_ref,  "\\n".join(self.content)[:120] + "..."
        )


class RestructuredSection:
    def __init__(self, title: str, heading_level: int, content: List[str]):
        """
        Represents a section of a reStructured document.

        Keyword Arguments:
        title: str         --
        heading_level: int --
        content: List[str] -- content of the section
        """
        self.title: List[str] = make_heading(title, heading_level)
        self.content: List[str] = content

    def is_empty(self) -> bool:
        return not self.content

    def as_text(self) -> List[str]:
        return self.title + self.content if not self.is_empty() else []


def wrap_in_newlines(restructured: List[str] = []) -> List[str]:
    return ["", *restructured, ""]


def make_heading(line: str, level: int = 1) ->List[str]:
    """
    reStructuredText headings are not well assigned but the convention
    followed is  
        # with overline, for parts

        \* with overline, for chapters
        
        = for sections
        
        - for subsections
        
        ^ for subsubsections
        
        " for paragraphs
    """

    level = 0 if level == 1 else 1

    heading_markers: List[str] = ["=", "-" ]
    heading_marker: str = heading_markers[level] * len(line)

    return ["", line, heading_marker, ""]


# def escape_markdown(text: str) -> str:
#     """Escapes characters that have a special meaning in markdown, like *_-"""
#     # characters: str = "*_-+`"
#     # for character in characters:
#     #     text = text.replace(character, '\\' + character)
#     # return text
#     return text.translate(str.maketrans({
#                                         "*": r"\*",
#                                         "_": r"\_",
#                                         "-": r"\-"
#                                         }))


def make_bold(text: str) -> str:
    """Returns the text surrounded by **"""
    return "**" + text + "**"


def make_italic(text: str) -> str:
    """Returns the text surrounded by *"""
    return "*" + text + "*"


def make_code_inline(text: str) -> str:
    """Returns the text surrounded by `"""
    return "`" + text + "`"


def make_code_block(text: str, language: str = "gdscript") -> str:
    """Returns the text formatted as a reStructured code block """
    lines: List[str] = text.split("\n")
    code: str = "    " + "\n    ".join(lines)
    return "..  code-block:: {}\n\n{}\n".format(language, code)


def make_link(description: str) -> str:
    if description == "var" or description == 'void':
        return description

    api_key: str = description.lower()
    if api_key in api_ref:
        LOGGER.info(
            "found ref, link is {}".format(api_ref[api_key])
        )
        return ":godot_class:`{} <{}>`".format(description, description.lower())
    else:
        link_target = "class_" + description
        return ":ref:`{} <{}>`".format(description, link_target)


def make_list(
    strings: List[str], is_numbered: bool = False, indent_level: int = 0
) -> List[str]:
    """Returns a bullet or ordered list from strings."""
    indent: str = "  " * indent_level

    def make_list_item(index: int, string: str) -> str:
        return indent + "{} {}".format(index + "." if is_numbered else "-", string)

    return [make_list_item(i, string) for i, string in enumerate(strings, start=1)]


def make_table_header(cells: List[str]) -> List[str]:
    return [make_table_row(cells), " --- |" * (len(cells) - 1) + " --- "]


def make_table_row(cells: List[str]) -> str:
    return " | ".join(cells)


def make_prop_table(props: List[str], class_name: str) -> List[str]:
    table_items: List[Dict] = []
    exp_length: int = 0
    type_length: int = 0
    name_length: int = 0
    default_length: int = 7
    for prop in props:
        prop_exported = "**export**" if prop.is_exported else ""
        prop_type: str = make_link(prop.type)
        prop_name: str = ":ref:`{}<class_{}_property_{}>`".format(
                                                            prop.name.lower(),
                                                            class_name,
                                                            prop.name.lower()
                                                        )
        prop_def: str = "``{}``".format(prop.default_value) if prop.default_value else ""
        table_items.append({"exported": prop_exported, "type": prop_type, "name": prop_name, "default": prop_def})
    
    for item in table_items:
        e_len = len(item["exported"]) +2
        t_len = len(item["type"]) + 2
        n_len = len(item["name"]) + 2
        d_len = len(item["default"]) + 2
        if e_len > exp_length: exp_length = e_len
        if t_len > type_length: type_length = t_len
        if n_len > name_length: name_length = n_len
        if d_len > default_length: default_length = d_len

    table_separator: str = "+{}+{}+{}+{}+".format(
        "-" * exp_length, "-" * name_length, "-" * type_length, "-" * default_length
        )
    
    table: List[str] =[]
    table.append(table_separator)
    for row in table_items:
        item_row: str = "| {} | {} | {} | {} |".format(
                row["exported"] + " " * (exp_length - len(row["exported"]) - 2),
                row["name"] + " " * (name_length - len(row["name"]) - 2),
                (row["type"] + " " * (type_length - len(row["type"]) - 2)) if row["type"] is not "var " else " var ",
                row["default"] + " " * (default_length - len(row["default"]) - 2)
            )

        table.append(item_row)
        table.append(table_separator)

    return table


def make_func_table(funcs: List[str], class_name: str) -> List[str]:
    table_items: List[Dict] = []
    ret_type_len: int = 0
    func_call_len: int = 0

    for func in funcs:
        ret_type: str = None
        if func.return_type == "null" or func.return_type == "void":
            ret_type = func.return_type
        else:
            ret_type = make_link(func.return_type)
        func_call: List[str] = []
        func_call.append(":ref:`{}<class_{}_method_{}>` **(** ".format(
                                                    func.name,
                                                    class_name,
                                                    func.name
                                                    )
                                                )
        if func.arguments:
            func_call.extend(make_arguments(func.arguments))
            # for arg in func.arguments:
            #     arg_type: str = None
            #     arg_default: str = None
            #     if arg.type == "null" or arg_type == "void":
            #         arg_type = arg.type
            #     else:
            #         arg_type = make_link(arg.type)
            #     arg_default = " = {}".format(arg.default) if arg.default != "" else arg.default
            #     argument: str = "{}: {}{}".format(arg.name, arg_type, arg_default)

            #     func_call.append(argument)
            #     func_call.append(", ")
            # func_call.pop() # get rid of the last comma        
        func_call.append(" **)**")

        table_items.append({"func_call": ''.join(func_call), "ret_type": ret_type})

    for item in table_items:
        c_len = len(item["func_call"]) +2
        r_len = len(item["ret_type"]) + 2
        if c_len > func_call_len: func_call_len = c_len
        if r_len > ret_type_len: ret_type_len = r_len

    table_separator: str = "+{}+{}+".format(
        "-" * func_call_len, "-" * ret_type_len
        )
    
    table: List[str] =[]
    table.append(table_separator)
    for row in table_items:
        item_row: str = "| {} | {} |".format(
                row["func_call"] + " " * (func_call_len - len(row["func_call"]) - 2),
                row["ret_type"] + " " * (ret_type_len - len(row["ret_type"]) - 2)
            )

        table.append(item_row)
        table.append(table_separator)

    return table


def make_element(attribute: str, element: List[str], class_name: str) -> List[str]:

    switcher = {
        "members": lambda : make_members(element, class_name),
        "functions": lambda : make_functions(element, class_name),
        "signals": lambda : make_signals(element, class_name),
        "enums": lambda : make_enums(element, class_name),
        "constants": lambda : make_constants(element, class_name),
    }
    return switcher.get(attribute, lambda: List("ERROR: attribute not known"))()


def make_members(members: List[str], class_name: str) -> List[str]:
    LOGGER.info(
        "Making members for class {}".format(class_name)
    )
    members_lines: List[str] = []
    for member in members:
        no_default: bool = True if member.default_value == None else False
        member_default: str = None if no_default else "``{}``".format(member.default_value)
        if member_default == '````':
            member_default = '*member has no default setting*'
        no_getter: bool = True if member.getter == '' else False
        member_getter: str = None if no_getter else "{}()".format(member.getter)
        no_setter: bool = True if member.setter == '' else False
        member_setter: str = None if no_setter else "{}(val)".format(member.setter)
        member_lines: List[str] = []
        member_lines.append('.. _class_{}_property_{}:\n'.format(class_name, member.name))
        member_lines.append('- **{}** : {}\n'.format(
                                            member.name,
                                            make_link(member.type)))
        
        if not (no_default and no_getter and no_setter):
            def_len = len(member_default) if member_default else 0
            set_len = len(member_setter) if member_setter else 0
            get_len = len(member_getter) if member_getter else 0

            table_arg_width = max(def_len, set_len, get_len)
            table_seperator = "+-----------+-{}-+".format('-' * table_arg_width)
            member_lines.append(table_seperator)
            if not no_default:
                member_lines.append("| *Default* | {} |".format(member_default + \
                    " " * (table_arg_width - len(member_default))))
                member_lines.append(table_seperator)
            if not no_setter:
                member_lines.append("| *Setter*  | {} |".format(member_setter + \
                    " " * (table_arg_width - len(member_setter))))
                member_lines.append(table_seperator)
            if not no_getter:
                member_lines.append("| *Getter*  | {} |".format(member_getter + \
                    " " * (table_arg_width - len(member_getter))))
                member_lines.append(table_seperator)
            member_lines.append('\n')
            

        member_lines.append('{}\n'.format(member.description))
        member_lines.append('----\n')

        members_lines.extend(member_lines)

    members_lines.pop()
    return members_lines


def make_functions(functions: List[str], class_name: str) -> List[str]:
    LOGGER.info(
        "Making functions for class {}".format(class_name)
    )
    function_lines: List[str] =[]
    for function in functions:
        function_lines.append('.. _class_{}_method_{}:\n'.format(class_name, function.name))
        function_line: List[str] = []
        function_line.append('- {} **{}(** '.format(function.kind.name, function.name))
        
        if function.arguments:
            function_line.extend(make_arguments(function.arguments))
        
        function_line.append(' **)**')
        
        if function.return_type:
            function_line.append(' -> {}'.format(make_link(function.return_type)))
        
        function_lines.append('{}\n'.format(''.join(function_line)))
        function_lines.append('{}\n'.format(function.description))
        function_lines.append('----\n')

    function_lines.pop()
    return function_lines

def make_signals(signals: List[str], class_name: str) -> List[str]:
    LOGGER.info(
        "Making signals for class {}".format(class_name)
    )

    signal_lines: List[str] =[]
    for signal in signals:
        signal_line: List[str] = []
        signal_line.append("_ **{}** **(** ".format( signal.name))
        signal_line.append('')
        signal_args: List[str] = []
        for argument in signal.arguments:
            signal_args.append("{}".format(argument))
            signal_args.append(", ")
        if signal_args: 
            signal_args.pop()
            signal_line.extend(signal_args)

        signal_line.append(" **)**")   
    
        signal_lines.append('.. _class_{}_signal_{}:'.format(class_name, signal.name))
        signal_lines.append('')
        signal_lines.append(''.join(signal_line))
        signal_lines.append('')
        signal_lines.append(signal.description)
        signal_lines.append('')
        signal_lines.append('____')

    signal_lines.pop() # remove last line separator

    return signal_lines


def make_enums(enumerations: List[str], class_name: str) -> List[str]:
    LOGGER.info(
        "Making enums for class {}".format(class_name)
    )
    enum_lines: List[str] = []
    for enums in enumerations:
        enum_lines.append('.. _enum_{}_{}:'.format(class_name, enums.name))
        enum_lines.append('')
        for val in enums.values:
            enum_lines.append('.. _class_{}_constant_{}:'.format(class_name, val))
            enum_lines.append('')
        enum_lines.append('enum **{}** :'.format(enums.name))
        enum_lines.append('')
        for enum in enums.values:
            enum_lines.append('- **{}** = **{}**'.format(enum, enums.values[enum]))
        enum_lines.append('\n{}'.format(enums.description))
        enum_lines.append('')
        enum_lines.append('----')
        enum_lines.append('')

    enum_lines.pop() # remove empty line an previious line serarator
    enum_lines.pop()

    return enum_lines  




def make_constants(constants: List[str], class_name: str) -> List[str]:
    LOGGER.info(
        "Making constants for class {}".format(class_name)
    )
    const_lines: List[str] = []
    for sigs in constants:
        const_lines.append('.. _class_{}_constant_{}:\n'.format(class_name, sigs.name))
    for const in constants:
       const_lines.append('- **{}** :{} = **{}** --- {}\n'.format(const.name, const.type, const.default_value, const.description.replace('\n', '')))

    return const_lines



def make_arguments(arguments: List[str]) -> List[str]:
    arg_list: list[str] = []
    for arg in arguments:
        arg_type: str = None
        arg_default: str = None
        # if "type" in arg:
        if arg.type == "null" or arg_type == "void":
            arg_type = arg.type
        else:
            arg_type = make_link(arg.type)

        # if "default" in arg: 
        arg_default = " = {}".format(arg.default) if arg.default != "" else arg.default
        
        argument: str = "{}: {}{}".format(arg.name, arg_type, arg_default)

        arg_list.append(argument)
        arg_list.append(", ")
    
    arg_list.pop() # get rid of the last comma before sending back
    return arg_list
    

def make_comment(text: str) -> str:
    """
    reSturctured text comment format -
    ..  This is a single line comment.

        And this continues the comment
    """
    lines: List[str] = text.split("\n")

    if len(lines) == 1:
        return "..  " + lines[0]
    else:
        return "..\n    " + "\n    ".join(lines)


def surround_with_html(text: str, tag: str) -> str:
    return "<{}>{}</{}>".format(tag, text, tag)



"""
Converts the json representation of GDScript 
classes as dictionaries into objects
"""

import itertools
from logging import debug
import operator
import re
from dataclasses import dataclass
from enum import Enum
from operator import itemgetter
from sys import version
from typing import List, Tuple
import logging

from .make_restructured import (make_bold, make_code_inline, make_list,
                                surround_with_html)
from .utils import build_re_pattern

BUILTIN_VIRTUAL_CALLBACKS = [
    "_process",
    "_physics_process",
    "_input",
    "_unhandled_input",
    "_gui_input",
    "_draw",
    "_get_configuration_warning",
    "_ready",
    "_enter_tree",
    "_exit_tree",
    "_get",
    "_get_property_list",
    "_notification",
    "_set",
    "_to_string",
    "_clips_input",
    "_get_minimum_size",
    "_gui_input",
    "_make_custom_tooltip",
]

TYPE_CONSTRUCTOR = "_init"

LOGGER = logging.getLogger()

@dataclass
class Metadata:
    """ Container for metadata for Elements """

    tags: List[str]
    category: str


def extract_metadata(description: str) -> Tuple[str, Metadata]:
    """
    Finds metadata keys in the provided description and returns the description
    without the corresponding lines, as we as the metadata.  In the source text,
    Metadata should be of the form key: value, e.g. category: CategoryNmae
    """

    tags: List[str] = []
    category: str = ""

    lines: List[str] = description.split("\n")
    description_trimmed: List[str] = []
    pattern_tags = build_re_pattern("tags")
    pattern_category = build_re_pattern("category")

    for _, line in enumerate(lines):
        line_stripped: str = line.strip().lower()

        match_tags = re.match(pattern_tags, line_stripped)
        match_category = re.match(pattern_category, line_stripped)
        if match_tags:
            tags = match_tags.group(1).split(",")
            tags = list(map(lambda t: t.strip(), tags))
        elif match_category:
            category = match_category.group(1)
        else:
            description_trimmed.append(re.sub('^ ', '', line).rstrip())
    
    metadata: Metadata = Metadata(tags, category)

    rebuilt_description: str = "\n".join(description_trimmed)
    description_double_breaks_only: str = re.sub(r"(?<!\n)\n(?!\n)", " ", rebuilt_description)

    return description_double_breaks_only, metadata

class FunctionTypes(Enum):
    METHOD = 1
    VIRTUAL = 2
    STATIC = 3


@dataclass
class ProjectInfo:
    name: str
    description: str
    version: str

    @staticmethod
    def from_dict(data: dict):
        return ProjectInfo(data["name"], data["description"], data["version"])


@dataclass
class Element:
    """ Base type for all main GDScript symbol types.  Contains properties common
    to Signals, Functions, Member variables, etc.
    """

    signature: str
    name: str
    description: str

    def __post_init__(self) -> str:
        _description, self.matadata = extract_metadata(self.description)
        self.description = _description.strip("\n")

    
    def get_heading_as_string(self) -> str:
        """
        Returns and empty string.  Virtual method to get a list of strings
        representing the element as a reStructured heading
        """
        return self.name


    def get_unique_attributes_as_restructured(self) -> List[str]:
        """
        Returns an empty list. Virtual method to get a list of strings 
        describing the unique attributes of this element.
        """
        return []

    
    @staticmethod
    def from_dict(data: dict) -> "Element":
        return Element(data["signature"], data["name"], data["description"])

@dataclass
class Signal(Element):
    arguments: List[str]

    @staticmethod
    def from_dict(data: dict) -> "Signal":
        return Signal(
            data["signature"], data["name"], data["description"], data["arguments"],
        )

@dataclass
class Argument:
    """
    Container for function arguments
    """

    name: str
    type: str
    default: str


@dataclass
class Function(Element):
    kind: FunctionTypes
    return_type: str
    arguments: List[Argument]
    rpc_mode: int

    def __post_init__(self):
        super().__post_init__()
        self.signature = self.signature.replace("-> null", "-> void", 1)
        self.return_type = self.return_type.replace("null", "void")

    def summarize(self) ->List[str]:
        return[self.return_type, self.signature]

    def get_heading_as_string(self) -> str:
        """
        Returns an empty list.  Virtual method to get a list
        of strings representing the element as a reStructured heading.
        """
        heading: str = self.name
        if self.kind == FunctionTypes.VIRTUAL:
            heading += " " + surround_with_html("(virtual)", "small")
        if self.kind == FunctionTypes.STATIC:
            heading += " " + surround_with_html("(static)", "small")
        return heading

    @staticmethod
    def from_dict(data: dict) -> "Function":
        kind: FunctionTypes = FunctionTypes.METHOD
        if data["is_static"]:
            kind = FunctionTypes.STATIC
        elif data["is_virtual"]:
            kind = FunctionTypes.VIRTUAL

        return Function(
            data["signature"],
            data["name"],
            data["description"],
            kind,
            data["return_type"],
            Function._get_arguments(data["arguments"]),
            data["rpc_mode"] if "rpc_mode" in data else 0,
        )

    @staticmethod
    def _get_arguments(data: List[dict]) -> List[Argument]:
        args: List[Argument] = []
        for entry in data:
            if not "default_value" in entry:
                entry["default_value"] = ""
            args.append(Argument(entry["name"], entry["type"], entry["default_value"]))
        return args

@dataclass
class Enumeration(Element):
    """
    Represents an enum with its constants.
    """

    values: dict

    @staticmethod
    def from_dict(data: dict) -> "Enumeration":
        return Enumeration(
            data["signature"], data["name"], data["description"], data["value"],
        )


@dataclass
class Member(Element):
    """
    Represents a property of member varialble
    """

    type: str
    default_value: str
    is_exported: bool
    setter: str
    getter: str

    def summarize(self) -> List[str]:
        return [self.type, self.name]

    def get_unique_attributes_as_restructured(self) -> List[str]:
        setget: List[str] = []
        if self.setter and not self.setter.startswith("_"):
            setget.append(make_bold("Setter") + ": " + make_code_inline(self.setter))
        if self.getter and not self.getter.startswith("_"):
            setget.append(make_bold("Getter") + ": " + make_code_inline(self.getter))
        setget = make_list(setget)
        if len(setget) > 0:
            setget.append("")
        return setget

    @staticmethod
    def from_dict(data: dict) -> "Member":
        return Member(
            data["signature"],
            data["name"],
            data["description"],
            data["data_type"],
            data["default_value"],
            data["export"],
            data["setter"],
            data["getter"],
        )


@dataclass
class Constant(Element):
    """
    Represents a constant
    """

    type: str
    default_value: str

    def summarize(self) -> List[str]:
        return [self.type, self.name]

    @staticmethod
    def from_dict(data: dict) -> "Constant":
        return Constant(
            data["signature"],
            data["name"],
            data["description"],
            data["data_type"],
            data["value"],
        )                                                                        


@dataclass
class GDScriptClass:
    name: str
    extends: str
    description: str
    path: str
    functions: List[Function]
    members: List[Member]
    constants: List[Constant]
    signals: List[Signal]
    enums: List[Enumeration]
    sub_classes: List["GDScriptClass"]

    def __post_init__(self):
        description, self.metadata = extract_metadata(self.description)
        self.description = description.strip("\n")

        elements = self.functions + self.members + self.signals + self.enums
        self.symbols = {element.name for element in elements}

    @staticmethod
    def from_dict(data: dict):
        # the extends_class field is a list in json even though it only
        # has one class
        extends: str = data["extends_class"][0] if data["extends_class"] else ""
        return GDScriptClass(
            data["name"],
            extends,
            data["description"],
            data["path"],
            _get_functions(data["methods"])
            + _get_functions(data["static_functions"], is_static=True),
            _get_members(data["members"]),
            _get_constants(data["constants"]),
            _get_signals(data["signals"]),
            [
                Enumeration.from_dict(entry)
                for entry in data["constants"]
                if entry["data_type"] == "Dictionary"
                and all(isinstance(v, int) for v in entry["value"].values()) 
            ],
            [GDScriptClass.from_dict(data) for data in data["sub_classes"]],
        )

    def get_extends_tree(self, classes: "GDScriptClasses") -> List[str]:
        """
        returns the list of ancestor classes, starting
        from self.extends.

        Arguments:

        - classes: a GDScriptClasses list of GDScriptClasses this object is part of.

        """
        extends: str = self.extends
        extends_tree: List[str] = []
        while extends != "":
            extends_tree.append(extends)
            extends = next((cls.extends for cls in classes if cls.name == extends), "")
        return extends_tree


class GDScriptClasses(list):
    """
    Container for a list of GDScriptClass objects

    Provides methods of filtering and grouping GDScript classes
    """

    def __init__(self, *args):
        super(GDScriptClasses, self).__init__(args[0])
        self.class_index = {
            gdscript_class.name: gdscript_class.symbols for gdscript_class in self
        }

    def _get_grouped_by(self, attribute: str) -> List[List[GDScriptClass]]:
        LOGGER.info(
         "GDScriptClasses._get_grouped_by checking: {}".format(type(self[0].__dict__["metadata"].category)),
         "GDScriptClassese._get_grouped_by attribuete: {}".format(attribute) 
        )
        # if not self or self[0].__dict__["metadata"].category == None:
        if not self or attribute not in self[0].__dict__:
            return []

        groups = []
        get_attribute = operator.attrgetter(attribute)
        data = sorted(self, key=get_attribute)
        for key, group in itertools.groupby(data, get_attribute):
            groups.append(list(group))
        return groups

    def get_grouped_by_category(self) -> List[List[GDScriptClass]]:
        """
        Returns a list of lists of GDScriptClass objects, 
        grouped by their `category` attribute.
        """
        return self._get_grouped_by("category")

    @staticmethod
    def from_dict_list(data: List[dict]):
        return GDScriptClasses(
            [GDScriptClass.from_dict(entry) for entry in data if entry["name"] != ""]
        )


def _get_signals(data: List[dict]) -> List[Signal]:
    return [Signal.from_dict(entry) for entry in data]


def _get_functions(data: List[dict], is_static: bool = False) -> List[Function]:
    """
    Returns a list of valid functions to put in the class reference.
    Skips built-in virtual callbacks, except for constructor functions marked for
    inclusion, and private methods.
    """
    functions: List[Function] = []
    for entry in data:
        name: str = entry["name"]
        if name in BUILTIN_VIRTUAL_CALLBACKS:
            continue
        if name == TYPE_CONSTRUCTOR and not entry["arguments"]:
            continue

        _, metadata = extract_metadata(entry["description"])

        is_virtual: bool = "virtual" in metadata.tags and not is_static
        is_private: bool = name.startswith(
            "-"
        ) and not is_virtual and name != TYPE_CONSTRUCTOR
        if is_private:
            continue

        function_data: dict = entry
        function_data["is_virtual"] = is_virtual
        function_data["is_static"] = is_static

        functions.append(Function.from_dict(function_data))
    return functions


def _get_members(data: List[dict]) -> List[Member]:
    return [
        Member.from_dict(entry) for entry in data if not entry["name"].startswith("-")
    ]


def _get_constants(constants_data: List[dict]) -> List[Constant]:
    """
    Filters and distinguishes constants from enums.
    """

    def is_not_enum(constant):
        """
        Returns `True` if the constant's source data doesn't correspond to an enum.
        That is, if it's not a dictionary with only a list of named integers.
        """
        
        is_enum = constant["data_type"] == "Dictionary"
        is_enum = is_enum and all(
            isinstance(value, int) for value in constant["value"].values()
        )
        return not is_enum

    constants = filter(lambda c: not c["name"].startswith("_"), constants_data)
    constants = filter(is_not_enum, constants)
    constants = sorted(constants, key = itemgetter("name"))
    return list(map(lambda c: Constant.from_dict(c), constants))




# import datetime
import sys
from argparse import ArgumentParser, Namespace


def parse(args=sys.argv) -> Namespace:
    parser: ArgumentParser = ArgumentParser(
        prog = "gdscript2rest",
        description="Converts JSON data dumped by Godot's "
        "GDScript language server to create .rst files for "
        "use with Sphinx.",
    )

    args.pop(0)  # we dont want the file list polluted with the program name.

    parser.add_argument("files", type=str, nargs="+", help="A list of paths to JSON files.")
    parser.add_argument("-p", "--path", type=str, default="export", help="Path to the output directory.")
    parser.add_argument("-i", "--make-index", action="store_true", default=False, help="If this flag is present, create and index.rst page with a table of contents.")
    parser.add_argument("-v", "--verbose", action="count", default=0, help="Set the verbosity level. For example -vv sets verbosity to level 2. Default: 0.")
    parser.add_argument("--dry-run", action="store_true", help="Run the script at max verbosity without creating files.  For debugging purposes.")
    parser.add_argument("-V", "--version", action="store_true", help="Print the version number and exit,")
    parser.add_argument("--doc-version", type=str, default="0.0.0", help="Set the document version number if there is no version set in the JSON file. Defaults to 0.0.0")

    namespace: Namespace = parser.parse_args(args)
    namespace.verbose = 9999 if namespace.dry_run else namespace.verbose
    return namespace
